/*
 * Copyright (c) 2007 Uwe Schmidt
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * 
 * Created on 09.03.2005
 */

package controller;

import java.awt.Graphics2D;
import java.awt.GraphicsConfiguration;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.Image;
import java.awt.image.BufferedImage;
import java.awt.image.ImageObserver;
import java.awt.image.PixelGrabber;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;

import javax.imageio.ImageIO;
import javax.swing.ImageIcon;
import javax.swing.JComponent;

import com.sun.image.codec.jpeg.JPEGCodec;
import com.sun.image.codec.jpeg.JPEGImageEncoder;

public class ImageHandler {

	private Model model;
	private Image img;
	private byte[][] pixel;
	private String[] decode;
	private int bytesToSend;
	private byte decodeByte;
	float huffp;

	// assign 256 gray values to four "led colors" of rotordisplay
	private final int sp0 = 0;
	private final int sp1 = 64;
	private final int sp2 = 128;
	private final int sp3 = 192;
	private final int sp4 = 255;

	// rotordisplay image size
	private int w = 200;
	private int h = 18;

	public ImageHandler(Model model) {
		this.model = model;
	}

	public void setImage(Image img) {
		if (this.img == img)
			return;
		this.img = img;
		w = img.getWidth(null);
		h = img.getHeight(null);

		model.log("Generating Image Data ... ", false, true);
		generateHuffmanCode();
		model.log("Done.", true, false);
	}

	public byte[][] getPixelData() {
		if (pixel == null) {
			model.log("Pixel data not generated yet.");
		}
		return pixel;
	}

	public byte getDecodeByte() {
		return decodeByte;
	}

	public String[] getCodingTable() {
		return decode;
	}

	/*
	 * generates byte-array out of the pixel color data of the image and the
	 * huffman coding table columns are shifted to the value given ('shift')
	 * shift can be negative ...
	 */
	public byte[] getImageData(byte[][] pixel, String[] decode, int bytesToSend, int shift) {

		// boolean usehuffman =
		// Integer.valueOf(model.getProperty("HuffmanCoding")).intValue()==1 ?
		// true : false;

		// temp.
		// if (!usehuffman) {
		// bytesToSend = 1000;
		// }

		byte[] out = new byte[bytesToSend];
		// out-array counter
		int oc = 0;
		// byte-string
		String tbyte = "";

		// loop over image pixel
		for (int i = 0; i < pixel.length; i++) {
			for (int j = 0; j < pixel[0].length; j++) {
				// calculate column index with shift
				int cIndex = (i + shift) % pixel.length;
				while (cIndex < 0)
					cIndex += pixel.length;

				// get code word for color of current pixel
				String t = decode[pixel[cIndex][j]];
				// add to byte-string (MSB is left)
				tbyte = t + tbyte;

				// temp: end of column => only 4 bits in this byte
				// if ( !usehuffman && (oc+1)%5==0 && tbyte.length()==4 ) {
				// tbyte = "0000" + tbyte;
				// }

				// if one byte is full
				if (tbyte.length() > 8) {
					// generate byte from byte-string
					out[oc++] = genByte(tbyte.substring(tbyte.length() - 8));
					// delete the substring that was converted to byte
					tbyte = tbyte.substring(0, tbyte.length() - 8);
				}
			}
		}

		// if we have "bits" left, append with zeros and generate byte
		if (tbyte.length() > 0) {
			out[oc++] = genByte("00000000".substring(0, 8 - tbyte.length()) + tbyte);
		}

		// if number of generates bytes differ with number of pre-calculated
		// bytes
		if (oc != bytesToSend) {
			model.log("Error while generating image data from huffman decoding table.");
			return null;
		}

		return out;
	}

	// generate a byte out of a string representing its binary notation
	private byte genByte(String s) {
		int ret = 0;
		for (int i = 0; i < 8; i++) {
			// add current bit
			ret += (byte) s.charAt(i) - '0';
			// left shift
			if (i != 7)
				ret = ret << 1;
		}
		return (byte) ret;
	}

	private void generateHuffmanCode() {

		// global pixel array, important for generating the data stream
		pixel = new byte[w][h];
		// decoding table
		decode = new String[4];
		// initialize counter array
		int[] c = new int[4];

		int[] pixels = new int[w * h];
		PixelGrabber pg = new PixelGrabber(img, 0, 0, w, h, pixels, 0, w);
		try {
			pg.grabPixels();
		} catch (InterruptedException e) {
			System.err.println("interrupted waiting for pixels!");
			return;
		}

		if ((pg.getStatus() & ImageObserver.ABORT) != 0) {
			System.err.println("image fetch aborted or errored");
			return;
		}

		c = new int[4];
		// loop over all pixels of the current image
		for (int i = 0; i < w; i++) {
			for (int j = 0; j < h; j++) {
				// determine "color"-value of pixel i,j
				pixel[i][j] = handlesinglepixel(pixels[j * w + i]);
				// inc color count
				c[pixel[i][j]]++;
			}
		}

		/*
		 * c[0] => #pixels of color 0 (black) c[1] => #pixels of color 1
		 * (darkgray) c[2] => #pixels of color 2 (lightgray) c[3] => #pixels of
		 * color 3 (white)
		 */

		boolean log = false;

		if (log)
			model.log("---------------------------------");
		if (log)
			model.log(pArr(c));

		// index array of minimums according to c
		int[] mindex = { 0, 1, 2, 3 };

		// ------------ sort mindex array according to the array 'c'
		for (int i = 0; i < c.length; i++) {
			int min = i;
			for (int j = i; j < c.length; j++) {
				if (c[mindex[j]] < c[mindex[min]])
					min = mindex[j];
			}
			int tmp = mindex[i];
			mindex[i] = mindex[min];
			mindex[min] = tmp;
		}
		// ------------ end of sorting

		/*
		 * mindex[0] => index of the color with fewest pixels in the image ...
		 * c[ mindex[0] ] => value of the minimum and so forth ...
		 */

		if (log)
			model.log("MinInd. " + pArr(mindex));

		boolean usehuffman = Integer.valueOf(model.getProperty("HuffmanCoding")).intValue() == 1 ? true : false;

		// use always the same coding-table (fixed 2-bit for each pixel)
		if (!usehuffman) {

			decode[0] = "00";
			decode[1] = "01";
			decode[2] = "10";
			decode[3] = "11";

			decodeByte = getHuffmanDecodeByte(2, 3, 1, 2, 0);

		} else {

			// READ decode-Table from right to left

			// if all colors are present in the image
			if (c[mindex[0]] != 0) {

				// 2 bits for all colors
				int sum01 = c[mindex[0]] + c[mindex[1]];
				if ((sum01 > c[mindex[2]]) && (sum01 > c[mindex[3]])) {
					decode[mindex[0]] = "00";
					decode[mindex[1]] = "10";
					decode[mindex[2]] = "01";
					decode[mindex[3]] = "11";

					// tree no.2
					decodeByte = getHuffmanDecodeByte(2, mindex[3], mindex[2], mindex[1], mindex[0]);

					// 1 bit for most occuring color, 2 for second and 3 for the
					// rest
				} else {
					decode[mindex[0]] = "000";
					decode[mindex[1]] = "100";
					decode[mindex[2]] = "10";
					decode[mindex[3]] = "1";

					// tree no.1
					decodeByte = getHuffmanDecodeByte(1, mindex[3], mindex[2], mindex[1], mindex[0]);

				}

				// if one color is missing
			} else if (c[mindex[1]] != 0) {

				if (log)
					model.log("Color " + mindex[0] + " is missing in the image.");
				decode[mindex[0]] = "";
				decode[mindex[1]] = "00";
				decode[mindex[2]] = "10";
				decode[mindex[3]] = "1";

				// tree no.3
				decodeByte = getHuffmanDecodeByte(3, mindex[3], mindex[2], mindex[1], 0);

				// two colors are missing
			} else {

				if (log)
					model.log("Color " + mindex[0] + " is missing in the image.");
				if (log)
					model.log("Color " + mindex[1] + " is missing in the image.");
				decode[mindex[0]] = "";
				decode[mindex[1]] = "";
				decode[mindex[2]] = "0";
				decode[mindex[3]] = "1";

				// tree no.4
				decodeByte = getHuffmanDecodeByte(4, mindex[3], mindex[2], 0, 0);
			}

		}

		// determine #bytes to be sent to the RD
		bytesToSend = (int) Math.ceil((c[0] * decode[0].length() + c[1] * decode[1].length() + c[2] * decode[2].length() + c[3] * decode[3].length()) / 8.0);

		if (log) {
			String s = "(reverse: ";
			for (int i = 0; i < decode.length; i++) {
				s += i + "='" + decode[i] + "', ";
			}
			s = s.substring(0, s.length() - 2) + ")";
			model.log(s);
			model.log("BytesToSend: " + bytesToSend);
		}

		int curbytes = 1000;

		int diff = curbytes - bytesToSend;
		if (log)
			model.log("Saving " + diff + " bytes (" + ((float) diff / curbytes) * 100 + "%)");
		if (log)
			model.log("DecodeByte = " + decodeByte);

		huffp = 1 - ((float) diff / curbytes);

		if (log)
			model.log("---------------------------------");
	}

	// // temp. to check 2 arrays for equality
	// private static boolean compare(byte[][] bit2data, byte[] huffdata) {
	// int p = 0;
	// for (int i=0; i<bit2data.length; i++) {
	// for (int j=0; j<bit2data[0].length; j++) {
	// if (bit2data[i][j] != huffdata[p++]) {
	// return false;
	// }
	// }
	// }
	// return true;
	// }

	public float getHuffPercentage() {
		return huffp;
	}

	public String pArr(int[] arr) {
		String s = "(";
		for (int i = 0; i < arr.length; i++) {
			s += arr[i] + ", ";
		}
		return s.substring(0, s.length() - 2) + ")";
	}

	public void createImage(String imagepath) {
		try {
			setImage(ImageIO.read(new File(imagepath)));
		} catch (IOException e) {
			model.log("Couldn't open image \"" + imagepath + "\".");
		}
	}

	public void createImage(JComponent comp) {

		GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
		GraphicsDevice gd = ge.getDefaultScreenDevice();
		GraphicsConfiguration gc = gd.getDefaultConfiguration();

		BufferedImage bImage = gc.createCompatibleImage(comp.getWidth(), comp.getHeight());

		Graphics2D g = bImage.createGraphics();
		comp.paint(g);

		setImage(bImage);
	}

	public byte[] getData() {
		return getData(0);
	}

	public byte[] getData(int shift) {
		return getImageData(pixel, decode, bytesToSend, shift);
	}

	// return an ImageIcon of the current bitmap
	public ImageIcon getImageIcon() {
		return new ImageIcon(img);
	}

	// ----------------------------------------------------------------------------

	// 2 bit color value for the rotordisplay leds
	public byte handlesinglepixel(int pixel) {

		// int alpha = (pixel >> 24) & 0xff;
		int red = (pixel >> 16) & 0xff;
		int green = (pixel >> 8) & 0xff;
		int blue = (pixel) & 0xff;

		int v = 0;

		// color image
		if (!((red == green) && (green == blue))) {
			// convert color pixel to luminance
			v = (int) Math.round(0.3 * red + 0.59 * green + 0.11 * blue);
			// grayscale image
		} else {
			v = red; // could also be green or blue, doesn't matter :)
		}

		// assign to predefined 4 intervals
		if (sp1 > v && v >= sp0)
			return 0;
		if (sp2 > v && v >= sp1)
			return 1;
		if (sp3 > v && v >= sp2)
			return 2;
		if (sp4 >= v && v >= sp3)
			return 3;

		return 0;
	}

	// save given image to file
	public void saveImage(Image img, String filename) {

		BufferedImage bimg = null;
		int w = img.getWidth(null);
		int h = img.getHeight(null);
		int[] pixels = new int[w * h];
		PixelGrabber pg = new PixelGrabber(img, 0, 0, w, h, pixels, 0, w);
		try {
			pg.grabPixels();
		} catch (InterruptedException ie) {
			ie.printStackTrace();
		}

		bimg = new BufferedImage(w, h, BufferedImage.TYPE_INT_RGB);
		bimg.setRGB(0, 0, w, h, pixels, 0, w);

		// Encode as JPEG
		try {
			FileOutputStream fos = new FileOutputStream(filename);
			JPEGImageEncoder jpeg = JPEGCodec.createJPEGEncoder(fos);
			jpeg.encode(bimg);
			fos.close();
		} catch (Exception ex) {
		}
	}

	/**
	 * returns a single byte that indicates the case for decoding the image data
	 * return value is 1-84
	 */
	private byte getHuffmanDecodeByte(int i, int j, int k, int l, int m) {
		byte[][][][][] dectable = new byte[4][4][4][4][4];
		dectable[0][0][1][2][3] = 1;
		dectable[0][0][1][3][2] = 2;
		dectable[0][0][2][1][3] = 3;
		dectable[0][0][2][3][1] = 4;
		dectable[0][0][3][1][2] = 5;
		dectable[0][0][3][2][1] = 6;
		dectable[0][1][0][2][3] = 7;
		dectable[0][1][0][3][2] = 8;
		dectable[0][1][2][0][3] = 9;
		dectable[0][1][2][3][0] = 10;
		dectable[0][1][3][0][2] = 11;
		dectable[0][1][3][2][0] = 12;
		dectable[0][2][0][1][3] = 13;
		dectable[0][2][0][3][1] = 14;
		dectable[0][2][1][0][3] = 15;
		dectable[0][2][1][3][0] = 16;
		dectable[0][2][3][0][1] = 17;
		dectable[0][2][3][1][0] = 18;
		dectable[0][3][0][1][2] = 19;
		dectable[0][3][0][2][1] = 20;
		dectable[0][3][1][0][2] = 21;
		dectable[0][3][1][2][0] = 22;
		dectable[0][3][2][0][1] = 23;
		dectable[0][3][2][1][0] = 24;
		dectable[1][0][1][2][3] = 25;
		dectable[1][0][1][3][2] = 26;
		dectable[1][0][2][1][3] = 27;
		dectable[1][0][2][3][1] = 28;
		dectable[1][0][3][1][2] = 29;
		dectable[1][0][3][2][1] = 30;
		dectable[1][1][0][2][3] = 31;
		dectable[1][1][0][3][2] = 32;
		dectable[1][1][2][0][3] = 33;
		dectable[1][1][2][3][0] = 34;
		dectable[1][1][3][0][2] = 35;
		dectable[1][1][3][2][0] = 36;
		dectable[1][2][0][1][3] = 37;
		dectable[1][2][0][3][1] = 38;
		dectable[1][2][1][0][3] = 39;
		dectable[1][2][1][3][0] = 40;
		dectable[1][2][3][0][1] = 41;
		dectable[1][2][3][1][0] = 42;
		dectable[1][3][0][1][2] = 43;
		dectable[1][3][0][2][1] = 44;
		dectable[1][3][1][0][2] = 45;
		dectable[1][3][1][2][0] = 46;
		dectable[1][3][2][0][1] = 47;
		dectable[1][3][2][1][0] = 48;
		dectable[2][0][1][2][0] = 49;
		dectable[2][0][1][3][0] = 50;
		dectable[2][0][2][1][0] = 51;
		dectable[2][0][2][3][0] = 52;
		dectable[2][0][3][1][0] = 53;
		dectable[2][0][3][2][0] = 54;
		dectable[2][1][0][2][0] = 55;
		dectable[2][1][0][3][0] = 56;
		dectable[2][1][2][0][0] = 57;
		dectable[2][1][2][3][0] = 58;
		dectable[2][1][3][0][0] = 59;
		dectable[2][1][3][2][0] = 60;
		dectable[2][2][0][1][0] = 61;
		dectable[2][2][0][3][0] = 62;
		dectable[2][2][1][0][0] = 63;
		dectable[2][2][1][3][0] = 64;
		dectable[2][2][3][0][0] = 65;
		dectable[2][2][3][1][0] = 66;
		dectable[2][3][0][1][0] = 67;
		dectable[2][3][0][2][0] = 68;
		dectable[2][3][1][0][0] = 69;
		dectable[2][3][1][2][0] = 70;
		dectable[2][3][2][0][0] = 71;
		dectable[2][3][2][1][0] = 72;
		dectable[3][0][1][0][0] = 73;
		dectable[3][0][2][0][0] = 74;
		dectable[3][0][3][0][0] = 75;
		dectable[3][1][0][0][0] = 76;
		dectable[3][1][2][0][0] = 77;
		dectable[3][1][3][0][0] = 78;
		dectable[3][2][0][0][0] = 79;
		dectable[3][2][1][0][0] = 80;
		dectable[3][2][3][0][0] = 81;
		dectable[3][3][0][0][0] = 82;
		dectable[3][3][1][0][0] = 83;
		dectable[3][3][2][0][0] = 84;
		return dectable[i - 1][j][k][l][m];
	}
}
