/*
 * Copyright (c) 2007 Uwe Schmidt
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * 
 * Created on 09.03.2005
 */

package controller;

import java.awt.FileDialog;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.io.File;
import java.io.FilenameFilter;
import java.util.HashMap;
import java.util.Map;

import javax.swing.JButton;
import javax.swing.JOptionPane;
import javax.swing.JTabbedPane;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import controller.gui.AboutWindow;
import controller.gui.MainWindow;
import controller.gui.PowerSlider;

public class InputController extends MouseAdapter implements ActionListener, ChangeListener, MouseWheelListener {

	public interface Command {
		public abstract void execute();
	}

	private Model model;
	private MainWindow frame;
	private Map<String, Command> commands;
	private AboutWindow about;

	// private StatusBar sb;

	public InputController() {
		commands = initCommandTable();
	}

	// Map ActionCommand to Command
	private Map<String, Command> initCommandTable() {
		Map<String, Command> tab = new HashMap<String, Command>();

		tab.put("Send Image", new Command() {
			public void execute() {
				model.log("Sending Image to RotorDisplay ...");
				model.createImage();
				model.sendImage();
			}
		});

		tab.put("Send Text", new Command() {
			public void execute() {
				model.log("Sending Text to RotorDisplay ...");
				model.log("Creating Image from Text.");
				model.createImage(frame.getEnterTextField());
				model.sendImage();
			}
		});

		tab.put("New", new Command() {
			public void execute() {
				model.getScriptPanel().clearFileAndText();
			}
		});

		tab.put("Open", new Command() {
			public void execute() {

				final int tabindex = Integer.parseInt(model.getProperty("TabIndex"));

				// Create File Dialog
				FileDialog fd = null;
				if (tabindex == Model.TAB_INDEX_IMAGE)
					fd = new FileDialog(frame, "Open Image");
				if (tabindex == Model.TAB_INDEX_SCRIPT)
					fd = new FileDialog(frame, "Open Script");
				if (fd == null)
					return;

				// Set initial directory
				fd.setDirectory(model.getProperty("LastDir"));
				fd.setLocationRelativeTo(frame);

				fd.setFilenameFilter(new FilenameFilter() {
					public boolean accept(File dir, String filename) {
						String filter = "";
						if (tabindex == Model.TAB_INDEX_IMAGE)
							filter = ".bmp";
						if (tabindex == Model.TAB_INDEX_SCRIPT)
							filter = ".txt";
						if (filename.indexOf(filter) != -1)
							return true;
						else
							return false;
					}
				});
				// Show File Dialog
				fd.setVisible(true);

				// User selected a File
				if (fd.getFile() != null) {

					// Update LastDir Property
					model.setProperty("LastDir", fd.getDirectory());

					if (tabindex == Model.TAB_INDEX_IMAGE) {
						// Set new ImagePath in Model
						String filepath = fd.getDirectory() + fd.getFile();
						model.log("Setting \"" + fd.getFile() + "\" as the new Image.");
						model.createImage(filepath);
					}

					if (tabindex == Model.TAB_INDEX_SCRIPT) {
						String filepath = fd.getDirectory() + fd.getFile();
						model.getScriptPanel().setFile(filepath);
					}
				}
			}
		});

		tab.put("Save", new Command() {
			public void execute() {

				if (model.getScriptPanel().fileOpened()) {
					model.getScriptPanel().saveFile();
				} else {
					// Create File Dialog
					FileDialog fd = null;
					fd = new FileDialog(frame, "Save Script");
					fd.setMode(FileDialog.SAVE);

					// Set initial directory
					fd.setDirectory(model.getProperty("LastDir"));
					fd.setLocationRelativeTo(frame);

					// Show File Dialog
					fd.setVisible(true);

					// User selected an File
					if (fd.getFile() != null) {

						// Update LastDir Property
						model.setProperty("LastDir", fd.getDirectory());

						String filepath = fd.getDirectory() + fd.getFile();
						model.getScriptPanel().saveFile(filepath);
					}
				}

			}
		});

		tab.put("Exit", new Command() {
			public void execute() {
				model.exitApp();
			}
		});

		tab.put("COM-Port", new Command() {
			public void execute() {
				// COM-Port Selection Dialog

				// get the currently active COM-Port from the Model
				String selectedElement = model.getProperty("COM");

				String s = (String) JOptionPane.showInputDialog(frame, "", "Select COM-Port", JOptionPane.PLAIN_MESSAGE, null,
				// get Portlist from model
						model.getPortList(), selectedElement);

				// User chose "Ok" => update the property in the model
				if ((s != null) && (s.length() > 0)) {
					// closing actual connection
					// model.closeConnection();
					// set new com-port
					model.log("Setting \"" + s + "\" as the new COM-Port.");
					model.setProperty("COM", s);
				}
			}
		});

		tab.put("Window", new Command() {
			public void execute() {
				Object[] options = { "save every time", "save current", "cancel" };
				int i = 0;
				if (model.getProperty("MainWindowSaveState").equals("no"))
					i = 1;

				int n = JOptionPane.showOptionDialog(frame, "Save window position and size every time\nthe window is closed or save current bounds?",
						"Window State", JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE, null, options, options[i]);

				// every time
				if (n == 0) {
					model.setProperty("MainWindowSaveState", "yes");
					model.log("Saving Window State every time.");
				}

				// keep current
				if (n == 1) {
					model.setProperty("MainWindowSaveState", "no");
					model.log("Saving Window State.");
					frame.saveMainWindowState();
				}

				// cancel is n=2 ..
			}
		});

		tab.put("About", new Command() {
			public void execute() {
				if (about == null)
					about = new AboutWindow(frame);
				about.setLocationRelativeTo(frame);
				about.setVisible(true);
			}
		});

		tab.put("L&F", new Command() {
			public void execute() {

				Map<String, String> lfs = new HashMap<String, String>();
				UIManager.LookAndFeelInfo[] arr = UIManager.getInstalledLookAndFeels();

				for (int i = 0; i < arr.length; i++) {
					lfs.put(arr[i].getName(), arr[i].getClassName());
				}

				// get currently acive L&F
				String selectedElement = UIManager.getLookAndFeel().getName();

				// why is the name different on mac os x?
				if (selectedElement.indexOf("Mac") != -1)
					selectedElement = selectedElement.replaceAll(" Aqua", "");

				String s = (String) JOptionPane.showInputDialog(frame, "", "Select L&F", JOptionPane.PLAIN_MESSAGE, null, lfs.keySet().toArray(),
						selectedElement);

				// User chose "Ok" => update the property in the model
				if ((s != null) && (s.length() > 0)) {
					// Set L&F
					try {
						model.log("Setting " + s + " Look&Feel ...");
						UIManager.setLookAndFeel(lfs.get(s).toString());
						SwingUtilities.updateComponentTreeUI(frame);
						model.setProperty("LookAndFeel", lfs.get(s).toString());
					} catch (Exception e) {
						model.log("Exception occurred: " + e);
					}

				}
			}
		});

		tab.put("Close Port", new Command() {
			public void execute() {
				model.closeConnection();
			}
		});

		tab.put("Install Comm-API", new Command() {
			public void execute() {
				CommInstaller ci = new CommInstaller(model);
				ci.installCommApi(frame);
			}
		});

		tab.put("Stop", new Command() {
			public void execute() {
				model.stopScript();
			}
		});

		/*
		 * tab.put("", new Command() { public void execute() { } });
		 */

		return tab;
	}

	public void setModel(Model model) {
		this.model = model;
	}

	public void setFrame(MainWindow frame) {
		this.frame = frame;
	}

	// public void setStatusbar(StatusBar sb) {
	// this.sb = sb;
	// }

	private void execCommand(String ac) {
		Command c = (Command) commands.get(ac);
		// if command exists
		if (c != null)
			c.execute();
		else {
			model.log("Command not implemented yet.");
		}

	}

	// Handle Mouse Clicks
	public void mouseClicked(MouseEvent e) {
		if (e.getComponent() instanceof JButton) {
			JButton button = (JButton) e.getComponent();
			if (button.isEnabled()) {
				execCommand(button.getActionCommand());
			}
			return;
		}

		if (e.getComponent() instanceof JTabbedPane) {
			JTabbedPane tp = (JTabbedPane) e.getComponent();
			model.setProperty("TabIndex", String.valueOf(tp.getSelectedIndex()));
			return;
		}
	}

	// Handle Actions
	public void actionPerformed(ActionEvent e) {
		// model.log(e+"");
		execCommand(e.getActionCommand());
	}

	// PowerSlider Change Event
	public void stateChanged(ChangeEvent e) {
		PowerSlider ps = (PowerSlider) e.getSource();
		if (!ps.getValueIsAdjusting()) {
			int value = (int) ps.getValue();
			model.setEnginePower(value);
		}
	}

	// PowerSlider Mouse Wheel Event
	public void mouseWheelMoved(MouseWheelEvent e) {
		// model.log(e+"");
		if (e.getComponent() instanceof PowerSlider) {
			PowerSlider ps = (PowerSlider) e.getComponent();
			int newval = ps.getValue() - e.getUnitsToScroll() / 3;
			ps.setValue(newval);
		}
	}

}
