/*
 * Copyright (c) 2007 Uwe Schmidt
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * 
 * Created on 08.03.2005
 */

package controller;

import java.awt.Image;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.Observable;
import java.util.Properties;

import javax.swing.ImageIcon;
import javax.swing.JComponent;
import javax.swing.JOptionPane;
import javax.swing.JTextField;
import javax.swing.UIManager;

import controller.gui.LogArea;
import controller.gui.MainWindow;
import controller.gui.PowerSlider;
import controller.gui.ScriptPanel;
import controller.script.RDScript;

public class Model extends Observable {

	// Messages sent to Oberservers
	public static final String IMAGE_CHANGE = "Image Change";
	public static final String IMAGE_VALID = "Image Valid";
	public static final String IMAGE_INVALID = "Image Invalid";
	public static final String PROPERTIES_CHANGE = "Prop Change";
	public static final String STATUS_CHANGE = "Status Change";
	public static final String EXIT = "Exit";
	public static final int TAB_INDEX_IMAGE = 0;
	public static final int TAB_INDEX_TEXT = 1;
	public static final int TAB_INDEX_SCRIPT = 2;

	// Main Attributes of the application
	private String imagepath = null;
	private Properties settings = null;
	private SettingsHandler sh = null;
	private String IniFile = "settings.ini";
	private ImageHandler imgh;
	private SerialConnection connection;
	private MainWindow gui;

	private LogArea log;
	private ScriptPanel sp;
	private RDScript parser;
	private PowerSlider ps;
	private JTextField enterTextField;

	// generate default settings for the application
	// in case the ini-file isn't found
	private Properties getDefaultSettings() {
		Properties p = new Properties();
		p.setProperty("COM", "COM1");
		p.setProperty("LastDir", ".");
		p.setProperty("MainWindowX", "-1");
		p.setProperty("MainWindowY", "-1");
		p.setProperty("MainWindowWidth", "600");
		p.setProperty("MainWindowHeight", "600");
		p.setProperty("MainWindowSaveState", "yes");
		p.setProperty("LookAndFeel", UIManager.getSystemLookAndFeelClassName());
		p.setProperty("ToolBarOrientation", "1");
		p.setProperty("Xon/Xoff", "1");
		p.setProperty("ConnectionSpeed", "115200");
		p.setProperty("ShiftSteps", "1");
		p.setProperty("ShiftWidth", "1");
		p.setProperty("OutBufferSize", "5");
		p.setProperty("log", "0");
		p.setProperty("Huffman", "0");
		p.setProperty("Direction", "left");
		p.setProperty("AckTimeout", "100");
		p.setProperty("HuffmanCoding", "0");
		p.setProperty("TabIndex", "0");
		return p;
	}

	public Model() {
		// Read Settings into an Property-Object (extends Hashtable)
		sh = new SettingsHandler(IniFile);
		// if Ini-File is not found, the method returns the passed default
		// properties
		settings = sh.getSettings(getDefaultSettings());
		// Bitmap-Handler
		imgh = new ImageHandler(this);
		// Connection with RotorDisplay
		connection = new SerialConnection(this);
	}

	public void saveImage(Image img, String filename) {
		imgh.saveImage(img, filename);
	}

	public void setLog(LogArea log) {
		this.log = log;
	}

	public void installCommApi() {

		// installation only supported on windows
		String os = System.getProperty("os.name");
		if (os.toLowerCase().indexOf("windows") == -1) {
			log("Please install the Comm-API.");
			return;
		}

		if (JOptionPane.showConfirmDialog(null, "Would you like to install the Comm-API?", "No Comm-API found ...", JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION) {
			log("Installing Comm-API ...");
			CommInstaller ci = new CommInstaller(this);
			ci.installCommApi(gui);
			JOptionPane.showMessageDialog(gui, "Please restart the application.");
			exitApp();
		}
	}

	// Log with Linefeed and Time-String
	public void log(String s) {
		log(s, true, true);
	}

	public void log(String s, boolean lf, boolean printTime) {
		GregorianCalendar cal = new GregorianCalendar();

		// must be an easier way...
		int ihour = cal.get(Calendar.HOUR_OF_DAY);
		int imin = cal.get(Calendar.MINUTE);
		int isec = cal.get(Calendar.SECOND);
		String hour = "00".substring(String.valueOf(ihour).length()) + ihour;
		String min = "00".substring(String.valueOf(imin).length()) + imin;
		String sec = "00".substring(String.valueOf(isec).length()) + isec;

		if (printTime)
			log.append(hour + ":" + min + ":" + sec + " ", false);
		log.append(s, lf);
	}

	// Exit Application
	public void exitApp() {
		log("Terminating ...");

		// set change flag
		setChanged();
		// notify oberservs that we're shutting down...
		notifyObservers(EXIT);

		sh.writeProperties(settings);
		System.exit(0);
	}

	public void setGui(MainWindow gui) {
		this.gui = gui;
	}

	// get a property
	public String getProperty(String key) {
		String res = settings.getProperty(key);
		// if property wasn't found, take default setting for that key
		if (res == null) {
			res = getDefaultSettings().getProperty(key);
			if (res == null)
				System.out.println("Property " + key + " wasn't found!");
			setProperty(key, res);
		}
		return res;
	}

	// set a property and update the ini-file
	public void setProperty(String key, String value) {
		settings.setProperty(key, value);
		// sh.writeProperties(settings);

		// set change flag
		setChanged();
		// notify oberservers that the properties have changed
		notifyObservers(PROPERTIES_CHANGE);
	}

	public void createImage(JComponent comp) {
		imgh.createImage(comp);
	}

	public void createImage() {
		createImage(imagepath);
	}

	public void createImage(String imagepath) {
		// set the new path
		this.imagepath = imagepath;
		imgh.createImage(imagepath);

		// set change flag
		setChanged();
		// notify oberservers that the image has changed
		notifyObservers(IMAGE_CHANGE);

		// set change flag
		setChanged();
		// notify oberservers that the image is now valid
		notifyObservers(IMAGE_VALID);
	}

	public String getImagePath() {
		return imagepath;
	}

	// convert unsigned byte to binary representation
	public static String toBinaryByte(byte b) {
		String s = "";
		for (int m = (b & 0xFF); m != 0; m /= 2) {
			s = (m % 2) + s;
		}
		return "00000000".substring(s.length()) + s;
	}

	// pass to ImageHandler Object
	public ImageIcon getImageIcon() {
		return imgh.getImageIcon();
	}

	// pass to SerialConnection
	public Object[] getPortList() {
		return connection.getPortList();
	}

	public void sendImage() {
		sendImage(0);
	}

	// send Image to RD -> pass to responsible object
	public void sendImage(int colStart) {
		if (connection.openPortAndListen()) {

			int times = Integer.parseInt(getProperty("ShiftSteps"));
			int width = Integer.parseInt(getProperty("ShiftWidth"));

			log("Starting to send ...");
			long dur_send = System.currentTimeMillis();

			int factor = getProperty("Direction").equals("left") ? 1 : -1;

			// send picture 'times' times width step width 'width'
			for (int i = 0; i < width * times; i += width)
				connection.writeImageData(imgh.getData(factor * i + colStart));

			dur_send = System.currentTimeMillis() - dur_send;
			log("Sending of " + times + " image(s) took " + dur_send + " msec (" + (times / (float) dur_send) * 1000 + " fps).");

			log("On average " + (float) dur_send / times + " msec/image.");
			// log("");
			// float newtimepp =
			// ((float)dur_send/times)*imgh.getHuffPercentage();
			// log("With huffman coding it would be " + newtimepp + "
			// msec/image.");
			// log("Overall time would then be " + newtimepp*times + " msec (" +
			// (times/(newtimepp*times))*1000 + " fps).");

		} else {
			log("Couldn't write image data!");
		}
	}

	public void closeConnection() {
		connection.closePort();
	}

	public SerialConnection getConnection() {
		return connection;
	}

	public void setEnginePower(int value) {
		connection.setEnginePower(value);
	}

	public void setCoilEnabled(boolean state) {
		connection.setCoilEnabled(state);
	}

	public void playScript() {
		if (parser == null) {
			String text = sp.getText();
			parser = RDScript.init(this, text);
		}
		parser.startThread();
	}

	public void pauseScript() {
		parser.suspendThread();
	}

	public void stopScript() {
		if (parser != null)
			parser.stopThread();
		setScriptRunning(false);
		parser = null;
	}

	public void setParser(RDScript parser) {
		this.parser = parser;
	}

	public void setScriptRunning(boolean state) {
		sp.setRunning(state);
	}

	public void setScriptPanel(ScriptPanel sp) {
		this.sp = sp;
	}

	public ScriptPanel getScriptPanel() {
		return sp;
	}

	public void setPowerSlider(PowerSlider ps) {
		this.ps = ps;
	}

	public PowerSlider getPowerSlider() {
		return ps;
	}

	public void setEnterTextField(JTextField enterTextField) {
		this.enterTextField = enterTextField;
	}

	public JTextField getEnterTextField() {
		return enterTextField;
	}

}
