/*
 * Copyright (c) 2007 Uwe Schmidt
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * 
 * Created on 09.03.2005
 */

package controller;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintStream;
import java.util.Enumeration;
import java.util.Map;
import java.util.Observable;
import java.util.Observer;
import java.util.SortedMap;
import java.util.TreeMap;

import javax.comm.CommPortIdentifier;
import javax.comm.PortInUseException;
import javax.comm.SerialPort;
import javax.comm.SerialPortEvent;
import javax.comm.SerialPortEventListener;
import javax.comm.UnsupportedCommOperationException;

public class SerialConnection extends Observable implements Observer, Runnable, SerialPortEventListener {

	public static final String ONLINE = "Online";
	public static final String OFFLINE = "Offline";

	private Map<String, CommPortIdentifier> portList;
	private CommPortIdentifier portId;
	private String comPort;
	private InputStream inputStream;
	private OutputStream outputStream;
	private SerialPort serialPort;
	// private Thread readThread;
	private Model model;
	private boolean online = false;

	public SerialConnection(Model model) {
		this.model = model;
		comPort = model.getProperty("COM");
		model.addObserver(this);
	}

	public boolean isOnline() {
		return online;
	}

	public void setOnline(boolean status) {
		online = status;

		setChanged();
		if (status) {
			notifyObservers(ONLINE);
		} else {
			notifyObservers(OFFLINE);
		}
	}

	public void update(Observable arg0, Object msg) {

		if (msg == Model.PROPERTIES_CHANGE) {
			// if com-port has changed ...
			if (!comPort.equals(model.getProperty("COM"))) {
				comPort = model.getProperty("COM");
				closePort();
			}
		}

		if (msg == Model.EXIT) {
			closePort();
		}
	}

	// returns the CommPortIdentifier of the selected Com-Port (Name taken from
	// Model)
	private CommPortIdentifier getPortId() {
		// generate PortList
		portList = generatePortlist();
		portId = portList.get(comPort);
		if (portId == null) {
			model.log("No valid Com-Port at \"" + comPort + "\"!");
		}
		return portId;
	}

	@SuppressWarnings("unchecked")
	private Map<String, CommPortIdentifier> generatePortlist() {
		// if portlist was generated previously
		if (portList != null)
			return portList;

		// else create new hashmap
		SortedMap<String, CommPortIdentifier> pList = new TreeMap<String, CommPortIdentifier>();

		// Enumeration of found COM-Ports
		Enumeration en = CommPortIdentifier.getPortIdentifiers();
		while (en.hasMoreElements()) {
			CommPortIdentifier portId = (CommPortIdentifier) en.nextElement();
			if (portId.getPortType() == CommPortIdentifier.PORT_SERIAL) {
				pList.put(portId.getName(), portId);
			}
		}

		if (pList.size() == 0) {
			model.log("No Com-Ports found!");
			model.installCommApi();
		}

		return pList;
	}

	public Object[] getPortList() {
		portList = generatePortlist();
		return portList.keySet().toArray();
	}

	// open the Serial Port and return a SerialPort-Object for further actions
	public SerialPort openSerialPort(CommPortIdentifier portId) {
		try {
			if (portId == null)
				return null;
			model.log("Opening Serial Port on \"" + portId.getName() + "\" ...");
			return (SerialPort) portId.open("RotorDisplay Controller", 2000);
		} catch (PortInUseException e) {
			model.log("Serial Port \"" + portId.getName() + "\" in use!");
			return null;
		}
	}

	// set the serial port parameters
	public void setSerialPortParams() {
		try {
			// get connection speed
			int speed = Integer.valueOf(model.getProperty("ConnectionSpeed")).intValue();
			model.log("Setting Serial Port Parameters ...");
			model.log(speed + " Bit, 8 Databit, 1 Stopbit, No Parity.");
			serialPort.setSerialPortParams(speed, SerialPort.DATABITS_8, SerialPort.STOPBITS_1, SerialPort.PARITY_NONE);

			// set outgoing flow control to XON/XOFF if set
			if (model.getProperty("Xon/Xoff").equals("1")) {
				model.log("XON/XOFF flow control on output.");
				serialPort.setFlowControlMode(SerialPort.FLOWCONTROL_XONXOFF_OUT);
			}

			// set OutBufferSize ...
			int bufsize = Integer.valueOf(model.getProperty("OutBufferSize")).intValue();
			serialPort.setOutputBufferSize(bufsize);
			model.log("Set OutputBufferSize to " + bufsize + ".");

			model.log("Done.");
		} catch (UnsupportedCommOperationException e) {
			model.log("Setting of Serial Port Parameters failed!");
		}
	}

	// set engine power level
	public void setEnginePower(int value) {

		if (!online) {
			if (!openPortAndListen())
				return;
		}

		// assert range 0-9
		if (value < 0 || value > 9) {
			model.log("Engine power out of range (" + value + ").");
			return;
		}

		try {
			// model.log("Setting engine power to level " + value + ".");
			outputStream.write(value + '0');
		} catch (IOException e) {
			model.log("Failed to set engine power to level " + value + ".");
		}

	}

	public void setCoilEnabled(boolean state) {

		if (!online) {
			if (!openPortAndListen())
				return;
		}

		try {
			if (state) {
				outputStream.write('c');
				model.log("enable coil: command 'c' ?");
			} else {
				outputStream.write('o');
				model.log("enable coil: command 'o' ?");
			}
		} catch (IOException e) {
			model.log("Failed to " + (state ? "enable" : "disable") + " coil.");
		}
	}

	// writes the image data array to the com-port
	public void writeImageData(byte[] data) {

		boolean log = Integer.valueOf(model.getProperty("log")).intValue() == 1 ? true : false;

		long duration = System.currentTimeMillis();

		if (log)
			model.log("Sending image data ...");

		// boolean out = false;

		try {

			// send 'b' to init transfer
			outputStream.write((byte) 'b');

			long dur_send = System.currentTimeMillis();

			// sending the whole image data ...
			outputStream.write(data);

			dur_send = System.currentTimeMillis() - dur_send;
			if (log)
				model.log("Sending took " + dur_send + " msec.");

			long dur_resp = System.currentTimeMillis();

			boolean recv_ok = false;

			int timeout = Integer.valueOf(model.getProperty("AckTimeout")).intValue();

			// loop since ok received or timeout (1000ms) occured ...
			while (!recv_ok && (System.currentTimeMillis() - dur_resp) < timeout) {

				// receive ok
				StringBuffer s = new StringBuffer();
				while (inputStream.available() > 0) {
					char recv = (char) inputStream.read();
					s.append(recv);
				}

				if (s.indexOf("OK") != -1) {
					recv_ok = true;
					if (log)
						model.log("Received OK after sending.");
				}

				if (s.indexOf("Failed") != -1) {
					if (log)
						model.log("Received Failed after sending.");
				}
				// recv end
			}

			dur_resp = System.currentTimeMillis() - dur_resp;
			if (log)
				model.log("Response took " + dur_resp + " msec.");

		} catch (IOException e) {
			model.log("Exception occured: " + e.getMessage());
		}

		if (log)
			model.log("Done.");

		// try {
		// Thread.sleep(2000); // Be sure data is xferred before closing
		// } catch (Exception e) {}

		duration = System.currentTimeMillis() - duration;
		if (log)
			model.log("sending and receiving took overall " + duration + " msec.");

	}

	public void closePort() {
		if (online) {
			model.log("Closing COM-Port on \"" + portId.getName() + "\" ...");
			serialPort.close();
		}
		// clear online flag
		setOnline(false);
	}

	// Lock-Step communication (serial, not event driven)
	public void sendBla() {
		try {
			PrintStream ps = new PrintStream(outputStream);
			ps.print("BLA");
			char c = (char) inputStream.read();
			System.out.println(c);

		} catch (IOException e) {
		}
	}

	// opens the serial port and adds a listener (Event-driven communication
	// needed?)
	public boolean openPortAndListen() {

		// if connection already set up
		if (online)
			return true;

		serialPort = openSerialPort(getPortId());
		if (serialPort == null)
			return false;

		try {
			inputStream = serialPort.getInputStream();
			outputStream = serialPort.getOutputStream();
		} catch (IOException e) {
			model.log("Failed to get In- and Output-Streams from Serial Port!");
		}

		// // add EventListener for reading
		// try {
		// serialPort.addEventListener(this);
		// } catch (TooManyListenersException e) {}
		//
		// serialPort.notifyOnDataAvailable(true);

		// set serial port parameters
		setSerialPortParams();

		// model.log("OutputBufferSize is " + serialPort.getOutputBufferSize() +
		// ".");

		// // start Thread
		// readThread = new Thread(this);
		// readThread.start();

		// set online flag
		setOnline(true);
		return true;
	}

	public void run() {
		try {
			Thread.sleep(20000);
		} catch (InterruptedException e) {
		}
	}

	public void serialEvent(SerialPortEvent event) {
		switch (event.getEventType()) {
			case SerialPortEvent.BI:
			case SerialPortEvent.OE:
			case SerialPortEvent.FE:
			case SerialPortEvent.PE:
			case SerialPortEvent.CD:
			case SerialPortEvent.CTS:
			case SerialPortEvent.DSR:
			case SerialPortEvent.RI:
			case SerialPortEvent.OUTPUT_BUFFER_EMPTY:
				break;

			case SerialPortEvent.DATA_AVAILABLE:
				byte[] readBuffer = new byte[1];
				try {
					while (inputStream.available() > 0) {
						// int numBytes = inputStream.read(readBuffer);
						inputStream.read(readBuffer);
					}
					// Log incoming data ... (temporary)
					// model.log("Received: " + new String(readBuffer));
				} catch (IOException e) {
					model.log(e.getMessage());
				}
				break;
		}
	}

}
