/*
 * Copyright (c) 2007 Uwe Schmidt
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * 
 * Created on 08.03.2005
 */

package controller.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.Observable;
import java.util.Observer;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JSplitPane;
import javax.swing.JTabbedPane;
import javax.swing.JTextField;
import javax.swing.UIManager;

import controller.InputController;
import controller.Model;

@SuppressWarnings("serial")
public class MainWindow extends JFrame implements Observer {

	private Container pane;
	private final Model model;
	private JButton wrapper;
	private JPanel bodyPane;
	private JButton sendImageButton, sendTextButton;
	private JTextField enterTextField;

	public JTextField getEnterTextField() {
		return enterTextField;
	}

	public void update(Observable arg0, Object msg) {
		// update the ImageIcon on the image preview

		// Image has changed
		if (msg == Model.IMAGE_CHANGE) {
			wrapper.setIcon(model.getImageIcon());
			wrapper.setDisabledIcon(model.getImageIcon());
			return;
		}
		// Image status is valid
		if (msg == Model.IMAGE_VALID) {
			sendImageButton.setEnabled(true);
		}

		// application exits
		if (msg.equals(Model.EXIT)) {
			// Save Window State if property is set
			if (model.getProperty("MainWindowSaveState").equals("yes")) {
				saveMainWindowState();
			}

		}
	}

	public void saveMainWindowState() {
		model.setProperty("MainWindowX", String.valueOf(getX()));
		model.setProperty("MainWindowY", String.valueOf(getY()));
		model.setProperty("MainWindowWidth", String.valueOf(getWidth()));
		model.setProperty("MainWindowHeight", String.valueOf(getHeight()));
	}

	public MainWindow(Model model, InputController ic) {

		// Set saved L&F
		try {
			UIManager.setLookAndFeel(model.getProperty("LookAndFeel"));
		} catch (Exception e) {
			model.log("Exception occurred: " + e);
		}

		this.model = model;
		model.addObserver(this);

		final Model m = model;
		this.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent arg0) {
				m.exitApp();
			}
		});

		// Set Title
		setTitle("RotorDisplay Control");

		// Set Bounds of the Window
		int x = Integer.valueOf(model.getProperty("MainWindowX")).intValue();
		int y = Integer.valueOf(model.getProperty("MainWindowY")).intValue();
		int w = Integer.valueOf(model.getProperty("MainWindowWidth")).intValue();
		int h = Integer.valueOf(model.getProperty("MainWindowHeight")).intValue();

		if (x == -1 && y == -1) {
			setBounds(0, 0, w, h);
			setLocationRelativeTo(null);
		} else {
			setBounds(x, y, w, h);
		}

		// Create MenuBar
		setJMenuBar(new GuiMenuBar(model, ic));

		// Add Items to the Panel
		pane = getContentPane();
		setDefaultLookAndFeelDecorated(true);
		pane.setLayout(new BorderLayout());

		// will be placed in SplitPane afterwards
		bodyPane = new JPanel(new BorderLayout());

		// Panel for Section "Image Send"
		JPanel sendImagePanel = new JPanel(new BorderLayout());

		// Wrapper-Button for the ImageIcon of the current Image
		wrapper = new JButton();
		wrapper.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
		wrapper.setFocusable(false);
		wrapper.setEnabled(false);
		sendImagePanel.add(wrapper, BorderLayout.CENTER);

		sendImageButton = new JButton("Send Image to RotorDisplay");
		sendImageButton.setActionCommand("Send Image");
		sendImageButton.setEnabled(false);
		// sendImageButton.addMouseListener(ic);
		sendImageButton.addActionListener(ic);
		sendImagePanel.add(sendImageButton, BorderLayout.SOUTH);
		// End of Section

		// Panel for Section "Text Send"
		JPanel sendTextPanel = new JPanel(new BorderLayout());

		// create Field to enter Text in the right dimensions
		enterTextField = new JTextField("RotorDisplay");
		enterTextField.setBorder(BorderFactory.createEtchedBorder());
		enterTextField.setActionCommand("Send Text");

		enterTextField.setPreferredSize(new Dimension(200, 18));
		enterTextField.setMinimumSize(new Dimension(200, 18));
		enterTextField.setMaximumSize(new Dimension(200, 18));
		enterTextField.setForeground(Color.WHITE);
		enterTextField.setBackground(Color.BLACK);
		enterTextField.setOpaque(true);
		// enterTextField.setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR));
		// enterTextField.setFont(new Font("", Font.PLAIN, 32));
		// enterTextField.setMargin(new Insets(0,0,0,0));
		enterTextField.addActionListener(ic);

		// Panel with FlowLayout required that the TextField keeps the size!
		JPanel flowContainer = new JPanel(new FlowLayout());
		flowContainer.add(enterTextField);
		// add Container to Panel
		sendTextPanel.add(flowContainer, BorderLayout.CENTER);

		// Send Text Button
		sendTextButton = new JButton("Send Text to RotorDisplay");
		sendTextButton.setActionCommand("Send Text");
		// sendTextButton.addMouseListener(ic);
		sendTextButton.addActionListener(ic);
		sendTextPanel.add(sendTextButton, BorderLayout.SOUTH);
		model.setEnterTextField(enterTextField);
		// End of Section

		// Panel for Section "Scripting"
		ScriptPanel sp = new ScriptPanel(model, ic);
		model.setScriptPanel(sp);
		// End of Section

		// TabbedPane for Sections
		JTabbedPane tabbedPane = new JTabbedPane();
		tabbedPane.addMouseListener(ic);
		tabbedPane.add(sendImagePanel, "Send Image");
		tabbedPane.add(sendTextPanel, "Send Text");
		tabbedPane.add(sp, "Scripting");

		tabbedPane.setSelectedIndex(Integer.parseInt(model.getProperty("TabIndex")));
		// add tabbedPane to bodyPane
		bodyPane.add(tabbedPane, BorderLayout.CENTER);

		// ToolBar
		String orientation = "";
		int or = Integer.valueOf(model.getProperty("ToolBarOrientation")).intValue();
		if (or == 0) {
			orientation = BorderLayout.NORTH;
		} else {
			orientation = BorderLayout.WEST;
		}
		GuiToolBar tb = new GuiToolBar(model, ic);
		tb.setOrientation(or);
		bodyPane.add(tb, orientation);

		// PowerSlider
		PowerSlider ps = new PowerSlider(model, ic);
		model.setPowerSlider(ps);
		bodyPane.add(ps, BorderLayout.EAST);

		// .....

		// Panel for the "SOUTH"-Region of the BorderLayout of the ContentPanel
		JPanel bottomPanel = new JPanel(new BorderLayout());

		StatusBar sb = new StatusBar(model);
		tb.setStatusbar(sb);
		bottomPanel.add(sb, BorderLayout.SOUTH);

		// Log-TextArea
		LogArea log = new LogArea(getBackground());
		// notify the model
		model.setLog(log);
		bottomPanel.add(log, BorderLayout.CENTER);

		// pane.add(bottomPanel, BorderLayout.SOUTH);

		// contains the 2 "big" panels
		JSplitPane splitPane = new JSplitPane(JSplitPane.VERTICAL_SPLIT, bodyPane, bottomPanel);
		splitPane.setOneTouchExpandable(true);
		splitPane.setDividerLocation(h * 6 / 10);

		// add to main panel
		pane.add(splitPane, BorderLayout.CENTER);

		// "hack" to notify components of activated tab (gray out images and
		// such)
		model.setProperty("TabIndex", String.valueOf(model.getProperty("TabIndex")));
	}
}
